import ij.*;
import ij.plugin.*;
import ij.gui.*;
import ij.process.*;

/*
Greyscale reconstruction, by G.Landini. 30/Oct/2003 for ImageJ

Reconstructs a greyscale image (the "mask" image) based on a "seed"
image.
This is an implementation of the parallel algorithm from:

Vincent L. Morphological greyscale reconstuction in Image Analysis:
Applications and efficient algorithms. IEEE Trans Image Proc 2(2)
176-201, 1993.

It is imperative to read Vincent's paper to understand greyscale
reconstruction and its applications.

The reconstruction algorithm is: iterated geodesic dilations of the
seed UNDER the mask image until stability is reached (the idempotent limit).

It supports stacks reconstructed by a single seed image and also
a single image reconstructed by a stack of seeds.

Plugin not tested in versions older than 1.31l

Send any improvements or bugs to G.Landini@bham.ac.uk

This plugin is based on Calculator_Plus by Wayne Rasband (wayne@codon.nih.gov)
*/

public class GreyscaleReconstruct_ implements PlugIn {

	static String title = "Greyscale Reconstruct";
    static boolean createWindow = true, whiteparticles = false;
	int[] wList;
	private String[] titles;
	int i1Index;
	int i2Index;
	ImagePlus i1;
	ImagePlus i2;
	boolean replicate;

	public void run(String arg) {
		if (IJ.versionLessThan("1.31l"))
			return;
		wList = WindowManager.getIDList();
		if (wList==null || wList.length<2) {
			IJ.showMessage(title, "There must be at least two windows open");
			return;
		}
		titles = new String[wList.length];
		for (int i=0; i<wList.length; i++) {
			ImagePlus imp = WindowManager.getImage(wList[i]);
			if (imp!=null)
				titles[i] = imp.getTitle();
			else
				titles[i] = "";
		}

		if (!showDialog())
			return;

		long start = System.currentTimeMillis();
		boolean calibrated = i1.getCalibration().calibrated() || i2.getCalibration().calibrated();

		if (calibrated)
			createWindow = true;
		if (createWindow) {
			if (replicate)
				i2 = replicateImage(i2, calibrated, i1.getStackSize());
			else
				i2 = duplicateImage(i2, calibrated);
			if (i2==null)
				{IJ.showMessage(title, "Out of memory"); return;}
			i2.show();
		}
		greyreconstruct(i1, i2);
		IJ.showStatus(IJ.d2s((System.currentTimeMillis()-start)/1000.0, 2)+" seconds");
	}

	public boolean showDialog() {
		GenericDialog gd = new GenericDialog(title);
		gd.addMessage("Note: Seed image should be\n\'darker\' than the mask image.");
		gd.addChoice("mask i1:", titles, titles[0]);
		gd.addChoice("seed i2:", titles, titles[1]);
		gd.addCheckbox("Create New Window", createWindow);
		gd.showDialog();
		if (gd.wasCanceled())
			return false;
		int i1Index = gd.getNextChoiceIndex();
		int i2Index = gd.getNextChoiceIndex();
		createWindow = gd.getNextBoolean();
		i1 = WindowManager.getImage(wList[i1Index]);
		i2 = WindowManager.getImage(wList[i2Index]);
		int d1 = i1.getStackSize();
		int d2 = i2.getStackSize();
		if (d2==1 && d1>1) {
			createWindow = true;
			replicate = true;
		}
		return true;
 }

	public void greyreconstruct(ImagePlus i1, ImagePlus i2) {
		int width  = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1, ip2;
		int slices1 = i1.getStackSize();
		int slices2 = i2.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		float[] ctable2 = i2.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		ImageStack stack2 = i2.getStack();
		int currentSlice = i2.getCurrentSlice();
		ImageStatistics stats;
		int  x,y,cp=0, pass=0;
		int [][] pixel = new int [width][height];
		int [][] pixel2 = new int [width][height];
		double prevsd, currsd, prevmean, currmean;
		IJ.showStatus("Greyscale Reconstruction...");
		for (int n=1; n<=slices2; n++) {
			ip1 = stack1.getProcessor(n<=slices1?n:slices1);
			stats=i1.getStatistics();

			ip2 = stack2.getProcessor(n);
			ip1.setCalibrationTable(ctable1);
			ip2.setCalibrationTable(ctable2);


			//get seed
			for(y=0;y<height;y++) {
				for(x=0;x<width;x++){
					pixel2[x][y]=ip2.getPixel(x,y);
					pixel[x][y]=pixel2[x][y];
				}
			}


			currsd = -1;
			currmean = -1;
			stats=i2.getStatistics();
			prevsd=stats.stdDev;
			prevmean=stats.mean;

			while (currsd!=prevsd && currmean!=prevmean){
				IJ.showStatus("Reconstructing: "+ ++pass);
				for (x=1; x<width-1; x++) {
					for (y=1; y<height-1; y++) {
						//white dilate of image except borders
						cp=0;
						if (pixel2[x-1][y-1]>cp) cp=pixel2[x-1][y-1];
						if (pixel2[x  ][y-1]>cp) cp=pixel2[x  ][y-1];
						if (pixel2[x+1][y-1]>cp) cp=pixel2[x+1][y-1];
						if (pixel2[x-1][y  ]>cp) cp=pixel2[x-1][y  ];
						if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
						if (pixel2[x+1][y  ]>cp) cp=pixel2[x+1][y  ];
						if (pixel2[x-1][y+1]>cp) cp=pixel2[x-1][y+1];
						if (pixel2[x  ][y+1]>cp) cp=pixel2[x  ][y+1];
						if (pixel2[x+1][y+1]>cp) cp=pixel2[x+1][y+1];
						pixel[x][y]=cp;
					}
				}

				y=0;
				for (x=1; x<width-1; x++) {
					//white dilate upper row
					cp=0;
					if (pixel2[x-1][y  ]>cp) cp=pixel2[x-1][y  ];
					if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
					if (pixel2[x+1][y  ]>cp) cp=pixel2[x+1][y  ];
					if (pixel2[x-1][y+1]>cp) cp=pixel2[x-1][y+1];
					if (pixel2[x  ][y+1]>cp) cp=pixel2[x  ][y+1];
					if (pixel2[x+1][y+1]>cp) cp=pixel2[x+1][y+1];
					pixel[x][y]=cp;
				}

				y=height-1;
				for (x=1; x<width-1; x++) {
					//white dilate lower row
					cp=0;
					if (pixel2[x-1][y-1]>cp) cp=pixel2[x-1][y-1];
					if (pixel2[x  ][y-1]>cp) cp=pixel2[x  ][y-1];
					if (pixel2[x+1][y-1]>cp) cp=pixel2[x+1][y-1];
					if (pixel2[x-1][y  ]>cp) cp=pixel2[x-1][y  ];
					if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
					if (pixel2[x+1][y  ]>cp) cp=pixel2[x+1][y  ];
					pixel[x][y]=cp;
				}

				x=0;
				for (y=1; y<height-1; y++) {
					//white dilate left column
					cp=0;
					if (pixel2[x  ][y-1]>cp) cp=pixel2[x  ][y-1];
					if (pixel2[x+1][y-1]>cp) cp=pixel2[x+1][y-1];
					if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
					if (pixel2[x+1][y  ]>cp) cp=pixel2[x+1][y  ];
					if (pixel2[x  ][y+1]>cp) cp=pixel2[x  ][y+1];
					if (pixel2[x+1][y+1]>cp) cp=pixel2[x+1][y+1];
					pixel[x][y]=cp;
				}

				x=width-1;
				for (y=1; y<height-1; y++) {
					//white dilate right column
					cp=0;
					if (pixel2[x-1][y-1]>cp) cp=pixel2[x-1][y-1];
					if (pixel2[x  ][y-1]>cp) cp=pixel2[x  ][y-1];
					if (pixel2[x-1][y  ]>cp) cp=pixel2[x-1][y  ];
					if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
					if (pixel2[x-1][y+1]>cp) cp=pixel2[x-1][y+1];
					if (pixel2[x  ][y+1]>cp) cp=pixel2[x  ][y+1];
					pixel[x][y]=cp;
				}

				x=0; //upper left corner
				y=0;
				cp=0;
				if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
				if (pixel2[x+1][y  ]>cp) cp=pixel2[x+1][y  ];
				if (pixel2[x  ][y+1]>cp) cp=pixel2[x  ][y+1];
				if (pixel2[x+1][y+1]>cp) cp=pixel2[x+1][y+1];
				pixel[x][y]=cp;

				x=width-1; //upper right corner
				//y=0;
				cp=0;
				if (pixel2[x-1][y  ]>cp) cp=pixel2[x-1][y  ];
				if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
				if (pixel2[x-1][y+1]>cp) cp=pixel2[x-1][y+1];
				if (pixel2[x  ][y+1]>cp) cp=pixel2[x  ][y+1];
				pixel[x][y]=cp;

				x=0; //lower left corner
				y=height-1;
				cp=0;
				if (pixel2[x  ][y-1]>cp) cp=pixel2[x  ][y-1];
				if (pixel2[x+1][y-1]>cp) cp=pixel2[x+1][y-1];
				if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
				if (pixel2[x+1][y  ]>cp) cp=pixel2[x+1][y  ];
				pixel[x][y]=cp;

				x=width-1; //lower right corner
				y=height-1;
				cp=0;
				if (pixel2[x-1][y-1]>cp) cp=pixel2[x-1][y-1];
				if (pixel2[x  ][y-1]>cp) cp=pixel2[x  ][y-1];
				if (pixel2[x-1][y  ]>cp) cp=pixel2[x-1][y  ];
				if (pixel2[x  ][y  ]>cp) cp=pixel2[x  ][y  ];
				pixel[x][y]=cp;

				//min of mask and dilated seed
				for (x=0; x<width; x++) {
					for (y=0; y<height; y++) {
						if(pixel[x][y]<ip1.getPixel(x,y)){
							pixel2[x][y]=pixel[x][y];
							ip2.putPixel(x,y, pixel[x][y]);
						}
						else{
							pixel2[x][y]=ip1.getPixel(x,y);
							ip2.putPixel(x,y, ip1.getPixel(x,y));
						}
					}
				}

				prevsd=currsd;
				prevmean=currmean;
				stats=i2.getStatistics();
				currsd=stats.stdDev;
				currmean=stats.mean;
			}


			if (n==currentSlice) {
				i2.getProcessor().resetMinAndMax();
				i2.updateAndDraw();
			}

			IJ.showProgress((double)n/slices2);
			IJ.showStatus(n+"/"+slices2);
		}
	}

	ImagePlus duplicateImage(ImagePlus img1, boolean calibrated) {
		ImageStack stack1 = img1.getStack();
		int n = stack1.getSize();
		ImageStack stack2 = img1.createEmptyStack();
		float[] ctable = img1.getCalibration().getCTable();
		try {
			for (int i=1; i<=n; i++) {
				ImageProcessor ip1 = stack1.getProcessor(i);
				ImageProcessor ip2 = ip1.duplicate();
				if (calibrated) {
					ip2.setCalibrationTable(ctable);
					ip2 = ip2.convertToFloat();
				}
				stack2.addSlice(stack1.getSliceLabel(i), ip2);
			}
		}
		catch(OutOfMemoryError e) {
			stack2.trim();
			stack2 = null;
			return null;
		}
		ImagePlus img2 =  new ImagePlus("Reconstructed", stack2);
		return img2;
	}

	ImagePlus replicateImage(ImagePlus img1, boolean calibrated, int n) {
		ImageProcessor ip1 = img1.getProcessor();
		ImageStack stack2 = img1.createEmptyStack();
		float[] ctable = img1.getCalibration().getCTable();
		try {
			for (int i=1; i<=n; i++) {
				ImageProcessor ip2 = ip1.duplicate();
				if (calibrated) {
					ip2.setCalibrationTable(ctable);
					ip2 = ip2.convertToFloat();
				}
				stack2.addSlice(null, ip2);
			}
		}
		catch(OutOfMemoryError e) {
			stack2.trim();
			stack2 = null;
			return null;
		}
		ImagePlus img2 =  new ImagePlus("Reconstructed", stack2);
		return img2;
	}
}

